function [costs, ws, omegas] = adaptiveSGD(data, labels, w, omega, pairing, gamma, maxIter, n0, C)
    costs = zeros(maxIter+1, 1);
    costs(1) = costfunction(data, labels, [w;omega], pairing, gamma);
    
    ws = zeros(maxIter+1, size(w, 1), size(w, 2));
    ws(1, :, :) = w;
    
    omegas = zeros(maxIter+1, size(omega, 1), size(omega, 2));
    omegas(1, :, :) = omega;
    
    [dim1 dim] = size(omega);
   
    %%Initialization
    M = [size(w, 1) + size(omega, 1), dim];
    g = zeros(M); %mean of gradient
    v = zeros(M); %variance gradient
    tau = n0*ones(M); %
   
    for i=1:n0
        grad = gradient(data(i,:), labels(i), w, omega, pairing(i, :), gamma);
        g = g + grad/n0;
        v = v + (grad.*grad)/n0;
    end
    nrProt = size(w, 1);
    h = abs((gradient(data(1:n0,:), labels(1:n0), w + g(1:nrProt, :), omega + g(nrProt+1:end, :), pairing(1:n0, :), gamma)  - g*n0)./(g*n0));
    h2 = h.*h;
    
    %Overestimate variance to decrease initial learning rate
    v = C*v;
    eta = (g.^2)./(h.*v);
    
    for iter=1:maxIter
        for i = 1:size(data,1) 
            grad= gradient(data(i,:), labels(i), w, omega, pairing(i, :), gamma);
            grad2 = gradient(data(i,:), labels(i), w + g(1:nrProt, :), omega + g(nrProt+1:end, :), pairing(i, :), gamma); %Evaluate gradient at theta+gradient
            curv = abs((grad2 - grad)./(g));
            
            g = (1 - (1./tau)).* g + (1./tau).*grad;
            v = (1 - (1./tau)).* v + (1./tau).*(grad.*grad);
            h = (1 - (1./tau)).* h + (1./tau).*curv;
            h2 = (1 - (1./tau)).* h2 + (1./tau).*curv.^2;
            
            eta = (g.^2)./(h.*v);
            
            w = w - eta(1:nrProt, :).*grad(1:nrProt, :);
            omega = omega  - eta(nrProt+1:end, :).* grad(nrProt+1:end, :);
            
            %Local Tau
            tau = (1 - g.^2./v).*tau + 1;
        end
        [costs(iter + 1), ~] = costfunction(data, labels, [w;omega], pairing, gamma);
        omegas(iter+1, :, :) = omega;
        ws(iter+1, :, :) = w;
    end
end

