%Input
%parameters: first rows are prototypes, then rows of tranformation matrix
%pairing: logical matrix specifying if prototype j has a similar label as 
%Output
%f: function value
%grad: gradient 
function [f, grad] = costfunction(data, labels, parameters, pairing, gamma)
    grad = zeros(size(parameters));
    nrOfProt = size(pairing, 2);
    w = parameters(1:nrOfProt, :);
    Omega = parameters(nrOfProt+1:end, :);
    
    %Determine distance matrix of data points to prototypes
    dist = pdist2(data*Omega', w*Omega');
    dist = dist.*dist;
    dist_correct = dist;
    dist_wrong = dist;

    %Set correct labels impossible
    dist_wrong(pairing) = inf; 
    dist_correct(~pairing) = inf;

    [distwrong idxwrong] = min(dist_wrong'); % Select closest wrong prototype
    [distcorrect idxcorrect] = min(dist_correct'); % Select closest correct prototype
    
    %Pre calculations
    distcorrectpluswrong = distcorrect + distwrong;
    distcorrectminuswrong = distcorrect - distwrong;
    callitq = distcorrectminuswrong ./ distcorrectpluswrong;
    callitq = 1./(1 + exp(-gamma * callitq)); % apply sigmoidal
    f=sum(callitq);
    
    deriv_sigmoid = gamma * callitq .* (1-callitq);
    normfactor = (distcorrect+distwrong).^2;
    lambda = Omega'*Omega;
    for i=1:size(data, 1)
        %Prototype gradients
        deltaJ = data(i, :) - w(idxcorrect(i), :);
        deltaK = data(i, :) - w(idxwrong(i), :);
        
        grad(idxcorrect(i), :) = grad(idxcorrect(i), :) - deriv_sigmoid(i)* (4 * distwrong(i)/normfactor(i) * lambda * deltaJ')';
        grad(idxwrong(i), :) = grad(idxwrong(i), :) + deriv_sigmoid(i)* (4 * distcorrect(i)/normfactor(i) * lambda * deltaK')';

        %Matrix gradients
        fJ = deriv_sigmoid(i)*(2*distwrong(i)/normfactor(i))*2*(Omega*deltaJ'*deltaJ);
        fK = deriv_sigmoid(i)*(2*distcorrect(i)/normfactor(i))*2*(Omega*deltaK'*deltaK);
        grad(nrOfProt+1:end, :) = grad(size(w, 1)+1:end, :) + fJ-fK;
    end
end