%%Generate two class data set 
dim=10;
N = 100;
w_center = [ones(1, dim)*1; ones(1, dim)*-1];

L = 2*rand(dim,dim) -1;

var1 = L*L';

%Generate data with multivariate normal distribution with mean w_i
data1 = mvnrnd(w_center(1,:), var1, N);
data2 = mvnrnd(w_center(2,:), var1, N);

data = [data1; data2];
labels = [ones(N, 1); ones(N, 1)*2]; 

%Intialize prototypes and transformation matrix
w = [mean(data1); mean(data2)] + rand(2, dim);
omega = 2*rand(dim, dim) - 1;

% Construct logical matrix with indicator if data point belongs to
% prototype
pairing = labels * [1, 2];
pairing = floor(sqrt(pairing)).^2 == pairing;

%Omega parameters
parameters = [w; omega];%concatenate prototype and matrix variables in 
gamma = 5; %Gamma parameters of sigmoidal function: 1/(1+exp(-gamma x))

%Shuffle data
I = randperm(size(data,1));
data = data(I, :);
labels = labels(I);
pairing = pairing(I, :);

%%% optimization options BFGS
% recommended: 'HessUpdate' -> {'lbfgs', 'bfgs', 'steepdesc'}
% recommended: 'Display' -> {'iter', 'off'}
% recommended: 'GradObj' -> {'on', 'off'} (see what happens if analytic gradient is ignored)
options = struct( ...
  'Display','iter', ...
  'GradObj','on', ...
  'GradConstr',false, ...
  'GoalsExactAchieve',0, ...
  'TolFun',1e-6, ...
  'MaxIter',100, ...
  'MaxFunEvals', 1000000, ...
  'TolX',1e-16, ...
  'DiffMinChange',1e-16, ...
  'HessUpdate','lbfgs', ...
  'storeN', 60 ...
);

[parameters,fvals] = fminlbfgs(@(parameters) costfunction(data, labels, parameters, pairing, gamma), parameters, options);

epochs = 100;
[costs1, w1, omega1] = adaptiveSGD(data, labels, w, omega, pairing, gamma, epochs, 200,2);
[costs2, w2, omega2] = adaptiveSGDb(data, labels, w, omega, pairing, gamma, epochs, 200,2);
[costs3, w3, omega3] = adaptiveSGDc(data, labels, w, omega, pairing, gamma, epochs, 200,2);

% %exhaustive search for learning rates
% c = [100; 10; 5; 1; 0.1; 0.05; 0.01; 0.005; 0.001; 5e-4; 1e-4];
% iter = 2;
% costs2 = zeros(numel(c)^2, iter+1);
% for i = 1:size(c, 1)
%     for j=1:size(c, 1)
%         eta = [ones(iter, 1)*c(i), ones(iter, 1)*c(j)];
%         [costs2((i-1)*size(c, 1)+j, :), ~, ~] = SGD(data, labels, w, omega, pairing, gamma, iter, eta);
%     end
% end
% 
% [~, ind] = min(costs2(:, iter+1));
% ind1 = ceil(ind/size(c, 1)); ind2 = mod(ind, size(c, 1));
% c(ind1)
% c(ind2)
% eta = [c(ind1)./(1 + 0.01*(1:epochs))', c(ind2)./(1 + 0.01*(1:epochs))'];
% [costs, w2, omega2] = SGD(data, labels, w, omega, pairing, gamma, epochs, eta);


% %%Generalization behavior?
% data1_new = mvnrnd(w_center(1,:), var1, N);
% data2_new = mvnrnd(w_center(2,:), var1, N);
% 
% data_new = [data1_new; data2_new];
% labels_new = [ones(N, 1); ones(N, 1)*2]; 
% 
% pairing_new = labels_new * [1, 2];
% pairing_new = floor(sqrt(pairing_new)).^2 == pairing_new;
% 
% gen = costfunction(data_new, labels_new, parameters, pairing_new, gamma)
% gen1 = costfunction(data_new, labels_new, [w1;omega1], pairing_new, gamma)
% %gen2 = costfunction(data_new, labels_new, [w2;omega2], pairing_new, gamma)



