%%Dataset settings
dim=25;
dim1= 15; %Projection to dim1-dimensional space
N = 50;

%Simulation settings
K=10; %Number of different datasets
P=5; %Number of initializations

%Storage of parameters/
data_cell = zeros(K, 2*N, dim);
labels_cell = zeros(K, 2*N);
L_cell = zeros(K, dim, dim);

w_inits = zeros(K, P, 2, dim);
omega_inits = zeros(K, P, dim1, dim);

parametersBFGS = cell(K, P);
parameterslBFGS = cell(K, P);

omegas_SGD = cell(K, P);
omegas_aSGD = cell(K, P);
omegas_aSGDb = cell(K, P);
omegas_aSGDc = cell(K, P);

ws_SGD = cell(K, P);
ws_aSGD = cell(K, P);
ws_aSGDb = cell(K, P);
ws_aSGDc = cell(K, P);

costs_BFGS = cell(K, P);
costs_lBFGS = cell(K, P);
costs_SGD = cell(K, P);
costs_aSGD = cell(K, P);
costs_aSGDb = cell(K, P);
costs_aSGDc = cell(K, P);

runtime_BFGS = zeros(K, P);
runtime_lBFGS = zeros(K, P);
runtime_aSGD = zeros(K, P);
runtime_aSGDb = zeros(K, P);
runtime_aSGDc = zeros(K, P);
runtime_grid = zeros(K, P);
runtime_SGD = zeros(K, P);

indi1 = zeros(K, P);
indi2 = zeros(K, P);

for k=1:K
    k
    w_center = [ones(1, dim)*1; ones(1, dim)*-1];

    L = 2*rand(dim,dim) -1;
    var1 = L*L';

    %Generate data with multivariate normal distribution with mean w_i
    data1 = mvnrnd(w_center(1,:), var1, N);
    data2 = mvnrnd(w_center(2,:), var1, N);

    data = [data1; data2];
    labels = [ones(N, 1); ones(N, 1)*2]; 

    % Construct logical matrix with indicator if data point belongs to
    % prototype
    pairing = labels * [1, 2];
    pairing = floor(sqrt(pairing)).^2 == pairing;

    %Shuffle data
    I = randperm(size(data,1));
    data = data(I, :);
    labels = labels(I);
    pairing = pairing(I, :);

    data_cell(k, :, :) = data;
    labels_cell(k, :) = labels;
    L_cell(k, :, :) = L;
    
    for p=1:P
        %Intialize prototypes and transformation matrix
        w = [mean(data1); mean(data2)] + (2*rand(2, dim) - 1);
        omega = 2*rand(dim1, dim) - 1;

        %Omega parameters
        parameters = [w; omega];%concatenate prototype and matrix variables in 
        gamma = 5; %Gamma parameters of sigmoidal function: 1/(1+exp(-gamma x))

        %%% optimization options BFGS
        % recommended: 'HessUpdate' -> {'lbfgs', 'bfgs', 'steepdesc'}
        % recommended: 'Display' -> {'iter', 'off'}
        % recommended: 'GradObj' -> {'on', 'off'} (see what happens if analytic gradient is ignored)
        options = struct( ...
          'Display','off', ...
          'GradObj','on', ...
          'GradConstr',false, ...
          'GoalsExactAchieve',0, ...
          'TolFun',1e-6, ...
          'MaxIter',100, ...
          'MaxFunEvals', 1000000, ...
          'TolX',1e-10, ...
          'DiffMinChange',1e-10, ...
          'HessUpdate','bfgs', ...
          'storeN', 60 ...
        );
        tic
        [parametersBFGS{k, p},costs_BFGS{k, p}] = fminlbfgs(@(parameters) costfunction(data, labels, parameters, pairing, gamma), parameters, options);
        runtime_BFGS(k, p) = toc;

        options = struct( ...
          'Display','off', ...
          'GradObj','on', ...
          'GradConstr',false, ...
          'GoalsExactAchieve',0, ...
          'TolFun',1e-6, ...
          'MaxIter',100, ...
          'MaxFunEvals', 1000000, ...
          'TolX',1e-10, ...
          'DiffMinChange',1e-10, ...
          'HessUpdate','lbfgs', ...
          'storeN', 20 ...
        );
    
        tic
        [parameterslBFGS{k, p},costs_lBFGS{k, p}] = fminlbfgs(@(parameters) costfunction(data, labels, parameters, pairing, gamma), parameters, options);
        runtime_lBFGS(k, p) = toc;
        
        epochs = 100;
        
        tic
        [costs_aSGD{k, p}, ws_aSGD{k, p}, omegas_aSGD{k, p}] = adaptiveSGD(data, labels, w, omega, pairing, gamma, epochs, 100,1);
        runtime_aSGD(k, p) = toc;
        
%         tic
%         [costs_aSGDb{k, p}, ws_aSGDb{k, p}, omegas_aSGDb{k, p}] = adaptiveSGDb(data, labels, w, omega, pairing, gamma, epochs, 100,1);
%         runtime_aSGDb(k, p) = toc;
%         
%         tic
%         [costs_aSGDc{k, p}, ws_aSGDc{k, p}, omegas_aSGDc{k, p}] = adaptiveSGDc(data, labels, w, omega, pairing, gamma, epochs, 100,1);
%         runtime_aSGDc(k, p) = toc;
        
        %exhaustive search for learning rates
        c = [100; 10; 5; 1; 0.1; 0.05; 0.01; 0.005; 0.001; 5e-4; 1e-4];
        iter = 2;
        costs2 = zeros(numel(c)^2, iter+1);
        tic
        for i = 1:size(c, 1)
            for j=1:size(c, 1)
                eta = [ones(iter, 1)*c(i), ones(iter, 1)*c(j)];
                [costs2((i-1)*size(c, 1)+j, :), ~, ~] = SGD(data, labels, w, omega, pairing, gamma, iter, eta);
            end
        end
        runtime_grid(k, p) = toc;
        

        [~, ind] = min(costs2(:, iter+1));
        ind1 = ceil(ind/size(c, 1)); ind2 = mod(ind, size(c, 1));
        eta = [c(ind1)./(1 + 0.01*(1:epochs))', c(ind2)./(1 + 0.01*(1:epochs))'];
        indi1(k, p) = ind1; indi2(k, p) = ind2;
        tic
        [costs_SGD{k, p}, ws_SGD{k, p}, omegas_SGD{k, p}] = SGD(data, labels, w, omega, pairing, gamma, epochs, eta);
        runtime_SGD(k, p) = toc;
    end
end
    


% %%Generalization behavior?
% data1_new = mvnrnd(w_center(1,:), var1, N);
% data2_new = mvnrnd(w_center(2,:), var1, N);
% 
% data_new = [data1_new; data2_new];
% labels_new = [ones(N, 1); ones(N, 1)*2]; 
% 
% pairing_new = labels_new * [1, 2];
% pairing_new = floor(sqrt(pairing_new)).^2 == pairing_new;
% 
% gen = costfunction(data_new, labels_new, parameters, pairing_new, gamma)
% gen1 = costfunction(data_new, labels_new, [w1;omega1], pairing_new, gamma)
% %gen2 = costfunction(data_new, labels_new, [w2;omega2], pairing_new, gamma)